/***************************************************************************
                          wdisplay.cpp  -  description
                             -------------------
    begin                : 14 11 2009
    copyright            : (C) 2009 by Andreas Lindenau
    email                : DL4JAL@darc.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <cmath>
#include <QtGui>

#include "wdisplay.h"

WDisplay::WDisplay(QWidget *parent) : QWidget(parent)
{
  //setBackgroundRole(QPalette::Dark);
  setAutoFillBackground(TRUE);
  setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
  setFocusPolicy(Qt::StrongFocus);
  rubberBandIsShown = FALSE;
  
  zoomInButton = new QPushButton("+",this);
//  zoomInButton->adjustSize();
  zoomInButton->setGeometry(0,0,20,20);
  connect(zoomInButton, SIGNAL(clicked()), this, SLOT(zoomIn()));
  
  zoomOutButton = new QPushButton("-",this);
  zoomOutButton->setGeometry(20,0,20,20);
  connect(zoomOutButton, SIGNAL(clicked()), this, SLOT(zoomOut()));
}

WDisplay::~WDisplay(){
  //zoomStack.clear();
}

void WDisplay::setPlotSettings(const PlotSettings &settings){
  zoomStack.clear();
  zoomStack.append(settings);
  curZoom = 0;
  zoomInButton->hide();
  zoomOutButton->hide();
  refreshPixmap();
}

void WDisplay::zoomOut(){
  if(curZoom > 0){
    --curZoom;
    zoomOutButton->setEnabled(curZoom > 0);
    zoomInButton->setEnabled(TRUE);
    zoomInButton->show();
    refreshPixmap();
  }
}

void WDisplay::zoomIn(){
  if(curZoom < zoomStack.count() -1){
    ++curZoom;
    zoomInButton->setEnabled(curZoom < zoomStack.count() -1);
    zoomOutButton->setEnabled(TRUE);
    zoomOutButton->show();
    refreshPixmap();
  }
}

void WDisplay::setCurveData(int id, const QVector<QPointF> &data, const QString &sugate){
  curveUgate[id] = sugate;
  curveMap[id] = data;
  refreshPixmap();
}

void WDisplay::clearCurve(int id){
  curveUgate.remove(id);
  curveMap.remove(id);
  refreshPixmap();
}

QSize WDisplay::minimumSizeHint() const{
  return QSize(6 * Margin, 4 * Margin);
}

QSize WDisplay::sizeHint() const{
  return QSize(12 * Margin, 8 * Margin);
}

void WDisplay::paintEvent(QPaintEvent *){
  QStylePainter painter(this);
  painter.drawPixmap(0,0,pixmap);
  if(rubberBandIsShown){
    painter.setPen(Qt::blue);
    //painter.setPen(palette().light().color());
    painter.drawRect(rubberBandRect.normalized().adjusted(0,0,-1,-1));
  }
  if(hasFocus()){
    QStyleOptionFocusRect option;
    option.initFrom(this);
    option.backgroundColor = palette().dark().color();
    painter.drawPrimitive(QStyle::PE_FrameFocusRect, option);
  }
}

void WDisplay::resizeEvent(QResizeEvent *event){
  refreshPixmap();
}

void WDisplay::mousePressEvent(QMouseEvent *event){
  QRect rect(Margin, Margin, width()-2*Margin, height()-2*Margin);
  if(event->button() == Qt::LeftButton){
    qDebug("mousePressEvent(QMouseEvent *event)");
    if(rect.contains(event->pos())){
      rubberBandIsShown = TRUE;
      rubberBandRect.setTopLeft(event->pos());
      rubberBandRect.setBottomRight(event->pos());
      updateRubberBandRegion();
      setCursor(Qt::CrossCursor);
    }
  }
}

void WDisplay::mouseMoveEvent(QMouseEvent *event){
  if(rubberBandIsShown){
    updateRubberBandRegion();
    rubberBandRect.setBottomRight(event->pos());
    updateRubberBandRegion();
  }
}

void WDisplay::mouseReleaseEvent(QMouseEvent *event){
  if((event->button() == Qt::LeftButton) && rubberBandIsShown){
    rubberBandIsShown = FALSE;
    updateRubberBandRegion();
    unsetCursor();
    
    QRect rect= rubberBandRect.normalized();
    if(rect.width() < 4 || rect.height() < 4) return;
    rect.translate(-Margin, -Margin);
    
    PlotSettings prevSettings = zoomStack[curZoom];
    PlotSettings settings = prevSettings;
    double dx = prevSettings.spanX() / (width() -2*Margin);
    double dy = prevSettings.spanY() / (height() -2*Margin);
    settings.minX = prevSettings.minX + dx * rect.left();
    settings.maxX = prevSettings.minX + dx * rect.right();
    settings.minY = prevSettings.maxY - dy * rect.bottom();
    settings.maxY = prevSettings.maxY - dy * rect.top();
    settings.adjust();
    
    zoomStack.resize(curZoom +1);
    zoomStack.append(settings);
    zoomIn();
  }
}

void WDisplay::keyPressEvent(QKeyEvent *event){
  switch(event->key()){
  case Qt::Key_Plus:
    zoomIn();
    break;
  case Qt::Key_Minus:
    zoomOut();
    break;
  case Qt::Key_Left:
    zoomStack[curZoom].scroll(-1,0);
    refreshPixmap();
    break;
  case Qt::Key_Right:
    zoomStack[curZoom].scroll(+1,0);
    refreshPixmap();
    break;
  case Qt::Key_Down:
    zoomStack[curZoom].scroll(0,-1);
    refreshPixmap();
    break;
  case Qt::Key_Up:
    zoomStack[curZoom].scroll(0,+1);
    refreshPixmap();
    break;
  default:
    QWidget::keyPressEvent(event);
  }
}

void WDisplay::wheelEvent(QWheelEvent *event){
  int numDegrees = event->delta() / 8;
  int numTicks = numDegrees / 15;
  
  if(event->orientation() == Qt::Horizontal){
    zoomStack[curZoom].scroll(numTicks,0);
  }else{
    zoomStack[curZoom].scroll(0,numTicks);
  }
  refreshPixmap();
}

void WDisplay::updateRubberBandRegion(){
  QRect rect=rubberBandRect.normalized();
  update(rect.left(), rect.top(), rect.width(), 1);
  update(rect.left(), rect.top(), 1, rect.height());
  update(rect.left(), rect.bottom(), rect.width(), 1);
  update(rect.right(), rect.top(), 1, rect.height());
}

void WDisplay::refreshPixmap(){
  PlotSettings settings = zoomStack[curZoom];
  
  pixmap = QPixmap(size());
  pixmap.fill(settings.colorhintergrund);
  QPainter painter(&pixmap);
  painter.initFrom(this);
  drawGrid(&painter);
  drawCurves(&painter);
  update();
}

void WDisplay::drawGrid(QPainter *painter){
  
  PlotSettings settings = zoomStack[curZoom];
  
  QRect rect(Margin, Margin, width() - 2 * Margin, height() - 2 * Margin);
  if(!rect.isValid())return;
  int yt;
  
  for(int i=0; i<= settings.numXTicks; ++i){
    int x=rect.left() + (i*(rect.width()-1)/settings.numXTicks);
    double label = settings.minX + (i*settings.spanX() / settings.numXTicks);
    painter->setPen(settings.pendotline);
    painter->drawLine(x, rect.top(), x, rect.bottom());
    painter->setPen(settings.pensolidline);
    painter->drawLine(x, rect.bottom(), x, rect.bottom() + 5);
    //Breite kliener 600 Pixel X-Beschriftung versetzt
    if(width() < 600){
      if(i%2){yt = 17;}else{yt = 5;}
    }else{
      yt = 5;
    }
    painter->drawText(x-50, rect.bottom()+yt,100,15,Qt::AlignHCenter | Qt::AlignTop, QString::number(label));
  }
  for(int j=0; j<= settings.numYTicks; ++j){
    int y=rect.bottom() - (j*(rect.height()-1)/settings.numYTicks);
    double label = settings.minY + (j*settings.spanY() / settings.numYTicks);
    painter->setPen(settings.pendotline);
    painter->drawLine(rect.left(), y, rect.right(), y);
    painter->setPen(settings.pensolidline);
    painter->drawLine(rect.left()-5, y, rect.left(), y);
    painter->drawText(rect.left()-Margin,y-10,Margin-5,20,Qt::AlignRight | Qt::AlignVCenter, QString::number(label));
  }
  //Y-Achse Bezeichnung
  painter->setPen(settings.pensolidline);
  painter->drawText(rect.left()-Margin/2-20,rect.top()-Margin/2,100,20,Qt::AlignLeft | Qt::AlignVCenter, settings.yeinheit);
  //X-Achse Bezeichnung
  painter->drawText(rect.right()-Margin/2-20,rect.bottom()+Margin/2,100,20,Qt::AlignLeft | Qt::AlignVCenter, settings.xeinheit);
  //3. Achse Bezeichnung
  painter->drawText(rect.right(),rect.top()-Margin/2,100,20,Qt::AlignLeft | Qt::AlignVCenter, settings.zeinheit);
  painter->drawRect(rect.adjusted(0,0,-1,-1));
  
  painter->setPen(settings.penueberschrift);
  painter->setFont( QFont("Sans", settings.ueberschriftsize));
  //Überschrift
  painter->drawText(rect.left(),rect.top()-Margin/2-20,rect.width(),25,Qt::AlignLeft | Qt::AlignVCenter, settings.ueberschrift);
}

void WDisplay::drawCurves(QPainter *painter){
  QString qs;
  double x, y, ym=0.0;
  
  PlotSettings settings = zoomStack[curZoom];
  
  QRect rectm(Margin, Margin, width() - 2 * Margin, height() - 2 * Margin);
  QRect rect(rectm);
  
  if(!rect.isValid())return;
  
  painter->setClipRect(rect.adjusted(1,1,-1,-1));
  painter->setFont( QFont("Sans", 10));
  QMapIterator<int, QVector<QPointF> > ih(curveMapH);
  QMapIterator<int, QString > jh(curveUgateH);
  while(ih.hasNext()){
    ih.next();
    int idh = ih.key();
    const QVector<QPointF> &datah = ih.value();
    QPolygonF polyline(datah.count());
    for(int kh=0; kh<datah.count(); ++kh){
      double dxh = datah[kh].x() - settings.minX;
      double dyh = datah[kh].y() - settings.minY;
      dyh *= -1.0;
      x= rect.left() + (dxh * (rect.width()-1) / settings.spanX());
      y= rect.bottom() + (dyh * (rect.height()-1) / settings.spanY());
      if(abs(rect.width()+Margin-x) < 10.0){
        ym = y;
      }
      polyline[kh]=QPointF(x,y);
      //qDebug("dx %f",dx);
    }
    //if(jh.hasNext()){
      jh.next();
      qs.sprintf("%i:",idh+1);
      qs = qs + jh.value();
    //}else{
    //  qs= "";
    //}
//    painter->setPen(colorForIds[uint(id) % 6]);
    painter->setPen(settings.penhlinie);
    painter->drawPolyline(polyline);
    //painter->setClipRect(rect.adjusted(0,0,Margin,0));
    painter->setClipRect(rect.adjusted(0,Margin*-1,Margin,0));
    //painter->setClipRect(rect.adjusted(0,0,Margin,0));
    if(ym < Margin)ym = Margin;
    painter->drawText(rect.width()+Margin+4,int(ym)-10,47,20,Qt::AlignLeft | Qt::AlignVCenter, qs);
    painter->setClipRect(rect);
  }
  
  rect = rectm;
  painter->setClipRect(rect.adjusted(1,1,-1,-1));
  QMapIterator<int, QVector<QPointF> > i(curveMap);
  QMapIterator<int, QString > j(curveUgate);
  while(i.hasNext()){
    i.next();
    int id = i.key();
    const QVector<QPointF> &data = i.value();
    QPolygonF polyline(data.count());
    for(int k=0; k<data.count(); ++k){
      double dx = data[k].x() - settings.minX;
      double dy = data[k].y() - settings.minY;
      dy *= -1.0;
      x= rect.left() + (dx * (rect.width()-1) / settings.spanX());
      y= rect.bottom() + (dy * (rect.height()-1) / settings.spanY());
      if(abs(rect.width()+Margin-x) < 10.0){
        ym = y;
      }
      polyline[k]=QPointF(x,y);
      //qDebug("dx %f",dx);
    }
    //if(j.hasNext()){
      j.next();
      qs.sprintf("%i:",id+1);
      qs = qs + j.value();
    //}else{
      //qs= "";
    //}
    painter->setPen(settings.penlinie);
    painter->drawPolyline(polyline);
    painter->setClipRect(rect.adjusted(0,Margin*-1,Margin,0));
    //painter->setClipRect(rect.adjusted(0,0,Margin,0));
    if(ym < Margin)ym = Margin;
    painter->drawText(rect.width()+Margin+4,int(ym)-10,47,20,Qt::AlignLeft | Qt::AlignVCenter, qs);
    painter->setClipRect(rect);
  }
}

void WDisplay::loeschenAlleHKurven(){
  curveUgateH.clear();
  curveMapH.clear();
  refreshPixmap();
}

void WDisplay::loeschenAlleKurven(){
  curveUgate.clear();
  curveMap.clear();
  refreshPixmap();
}

void WDisplay::Kurven2Hintergrund(){
  curveUgateH = curveUgate;
  curveMapH = curveMap;
  refreshPixmap();
}

PlotSettings::PlotSettings(){
  /*
  minX = 0.0;
  maxX = 10.0;
  numXTicks = 10;
  
  minY = 0.0;
  maxY = 500.0;
  numYTicks = 10;
  xeinheit = "Uds (V)";
  yeinheit = "Ids (mA)";
  zeinheit = "Ugs (V)";
  ueberschrift = "Ausgangskennlinie";
  */
}

void PlotSettings::setminX(double d){
  minX = d;
}

void PlotSettings::setmaxX(double d){
  maxX = d;
}

void PlotSettings::setminY(double d){
  minY = d;
}

void PlotSettings::setmaxY(double d){
  maxY = d;
}

void PlotSettings::setnumXTicks(int i){
  numXTicks = i;
}

void PlotSettings::setnumYTicks(int i){
  numYTicks = i;
}

void PlotSettings::setxeinheit(const QString &s){
  xeinheit = s;
}

void PlotSettings::setyeinheit(const QString &s){
  yeinheit = s;
}

void PlotSettings::setzeinheit(const QString &s){
  zeinheit = s;
}

void PlotSettings::setueberschrift(const QString &s){
  ueberschrift = s;
}

void PlotSettings::scroll(int dx, int dy){
  
  double stepX = spanX() / numXTicks;
  minX += dx * stepX;
  maxX += dx * stepX;
  
  double stepY = spanY() / numYTicks;
  minY += dy * stepY;
  maxY += dy * stepY;
}

void PlotSettings::adjust(){
  
  adjustAxis(minX, maxX, numXTicks);
  adjustAxis(minY, maxY, numYTicks);
}

void PlotSettings::adjustAxis(double &min, double &max, int &numTicks){
  
  const int MinTicks = 4;
  double grossStep = (max - min) / MinTicks;
  double step = pow(10.0, floor(log10(grossStep)));
  
  if(5*step < grossStep){
    step *= 5;
  }else{
    if(2*step < grossStep)step *= 2;
  }
  
  numTicks = int(ceil(max/step) - floor(min/step));
  if(numTicks < MinTicks)numTicks = MinTicks;
  min = floor(min/step)*step;
  max = ceil(max/step)*step;
}

void PlotSettings::setcolorhintergrund(QColor c){
  colorhintergrund = c;
}

void PlotSettings::setcolorschrift(QColor c){
  penschrift = QPen( c, 0, Qt::SolidLine);
  pensolidline = QPen( c, 0, Qt::SolidLine);
  pendotline = QPen( c, 0, Qt::DotLine);
}

void PlotSettings::setcolorueberschrift(QColor c, int size){
  penueberschrift = QPen( c, 0, Qt::SolidLine);
  ueberschriftsize = size;
}

void PlotSettings::setcolorlinie(QColor c){
  penlinie = QPen( c, 0, Qt::SolidLine);
}

void PlotSettings::setcolorhlinie(QColor c){
  penhlinie = QPen( c, 0, Qt::SolidLine);
}
